/*
 * Copyright (c) 2012, Willow Garage, Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *     * Neither the name of the Willow Garage, Inc. nor the names of its
 *       contributors may be used to endorse or promote products derived from
 *       this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */
#ifndef MYVIZ_H
#define MYVIZ_H


#include <QWidget>
#include<QString>
#include <QGridLayout>
#include<ros/ros.h>
#include"std_msgs/String.h"
#include<thread>
#include <visualization_msgs/Marker.h>

#include<QTimer>
#include <QMutex>
#include <QProcess>
#include <QThread>
#include<QSlider>
#include <rosbag/bag.h>
#include <sensor_msgs/LaserScan.h>
#include <rosbag/view.h>

namespace rviz
{
class Display;
class RenderPanel;
class VisualizationManager;
class ToolManager;
class PropertyTreeWidget;
}

// BEGIN_TUTORIAL
// Class "MyViz" implements the top level widget for this example.
class MyViz: public QWidget
{
Q_OBJECT
public:
  MyViz( QWidget* parent = 0 );
  virtual ~MyViz();
  void subCallback(const std_msgs::String& msg);
  void pubThread();
  void PlayCircle(int index);  // 圆圈线程执行函数
  rosbag::View *view;
private Q_SLOTS:   // QT信号和槽
  void setThickness( int thickness_percent );
  void setCellSize( int cell_size_percent );
  void setCloudTopic(const QString &newTopic);
  void setLaserTopic(const QString &newTopic);
  void setCloudSize(int cloudsize);
  void setLaserSize(int lasersize);

  void LaserColorChanged(int index); // 雷达颜色
  void GridColorChanged(int index); // 格栅颜色
  void setFixedFrame(const QString &FixedFrame);
  void setBackgroundColor(const QString& bg_color); // 背景颜色
  void AxesDisplayChanged(int index); // axe
  void CircleDisplayChanged(int index); // circle
  void Tree_Display(QGridLayout* layout, int index);
  
  void slot_select(); // 切换至select模式
  void slot_move_camera();

  // off
  double extractDuration(const QByteArray& output);
  void selectFile();
  void SpeedChanged(int index);
  void openFile();
private:
  rviz::VisualizationManager* manager_; // 
  rviz::RenderPanel* render_panel_;
  rviz::ToolManager* tool_manager_;
  rviz::Display* grid_;
  rviz::Display* cloud;
  rviz::Display* laser_;
  rviz::Display* axes_display;
  rviz::Display* marker_;
  rviz::PropertyTreeWidget* tree_widget_ ;
  ros::NodeHandle nh;
  ros::Subscriber sub;
  ros::Publisher pub;
  std::thread pub_thread;
  std::vector<QString> color_name; // 颜色的名称集合
  QString color_icon_filename;
  // off
  bool isclicked;// 用于判断play第几次按下play键
  bool ispauseclick;// 用于判断暂停键第几次按下
  int count;// 用于记录timeout的次数
  bool sliderreleased;// 用于判断是否拖动了进度条
  bool playagain;// 用于检测是否多次按下play按钮
  bool isthreadplaying;// 用于标识线程是否在运行
  bool abs;// 用于单独开线程？
  bool stored;// 用于存储一次容器,只在第一次play的时候存储容器
  bool is_pause = false;
  bool is_reset = false;
  bool is_end = false;
  bool is_save = false;
  bool is_saveMul = false;
  bool is_bagClosed = true;
  bool is_rateChanged = false;
  bool is_openFile = false;
  bool is_setFrameIndex = false;
  bool is_saveMulFinish = false;
  bool end = false;
  bool threadout = false;
  bool ischanged = false;
  bool ismoved = false;
  double speed_rate;
  QSlider* progress;
  QString filePath;
  double duration;
  int remain;
  int prog_num = 0;
  std::vector<QString> speed_name;
  ros::Publisher scan_pub;
  std::vector<sensor_msgs::LaserScan::ConstPtr> indx;// 存储节点信息
  int countbag;// 存储播放的具体位置
  double indxsize;
  int test = 0;


};
// END_TUTORIAL
#endif // MYVIZ_H
